#ifndef _AUDIOTAG_H_
#define _AUDIOTAG_H_

#include <stdint.h>

#define R3(x, A, B, C, D) \
    ((double)(C) + ((double)(x)-(double)(A)) * ((double)(D)-(double)(C)) \
                                             / ((double)(B)-(double)(A)))

#define ADD(var, val) __sync_fetch_and_add(&(var), val)
#define GET(var) __sync_fetch_and_add(&(var), 0)
#define SET(var, val) (void)__sync_lock_test_and_set(&(var), (val))

typedef struct {
  float min;
  float max;
} chunk;

typedef struct {
  uint64_t frames;
  int channels;
  int samplerate;
  chunk *chunks;
} input;

typedef struct {
  float *window_min;
  float *window_max;
  float *all_min;
  float *all_max;
} channel;

typedef struct {
  char *t;
  int tlen;
  int x;
  int y;
  int width;
  int height;
  int active;
} button;

typedef struct {
  int width;
  int height;
  int64_t samples_per_pixel;
  int64_t start_sample;
  channel *c;
  char *colors[12];
  int refresh_rate; /* min: 1Hz, max: what you want, 60 suggested */
  int64_t head;
  int64_t tail;
  char *font;
  button bt[5];
  int bt_count;
  char *region_filename;
  int no_follow;
  float vzoom;
  int no_bars;
} gui;

typedef struct {
  char *audio_file;
  int64_t start;
  int64_t current;
  int in_play;
  float buf[32768];
  float *read_buf;
  int buf_next;
  int free_samples;
  int free_pos;
  void *jack_port;
  int jack_error;
  int no_jack;
} player;

void *open_audio_file(char *, void *);
void minmax_fast(input *, uint64_t, chunk *);
void minmax(input *, uint64_t, uint64_t, chunk *, void *, float *);
void run_gui(gui *, input *, player *);
void init_jack(player *);
void start_play(player *);
void stop_play(player *);
void load_chunk(input *, char *, char *);
void run_compute(gui *g, input *r, player *p);

extern int gui_read, gui_write;
extern int slow_compute_lo_read, slow_compute_lo_write;
extern int fast_compute_lo_read, fast_compute_lo_write;
extern int slow_compute_hi_read, slow_compute_hi_write;
extern int fast_compute_hi_read, fast_compute_hi_write;
extern int jack_player_read, jack_player_write;
extern int jack_process_read, jack_process_write;
extern int compute_lo_timer_read, compute_lo_timer_write;

extern float timer_length;

extern volatile int hi_turn;
extern volatile int lo_turn;

void init_com(void);
void com_jack_player_started(void);
void com_jack_player_advance(void);
void com_jack_player_ended(void);
void com_jack_player_start(void);
void com_jack_player_stop(void);
void com_chunk_redraw(void);
void com_chunk_done(void);
void com_compute_lo_responsive(void);
void com_redraw_gui_now(void);

#endif /* _AUDIOTAG_H_ */
