#include "audiotag.h"
#include <stdlib.h>
#include <pthread.h>
#include <sndfile.h>
#include <unistd.h>

typedef struct {
  char *filename;
  input *r;
  volatile int run;
} chunk_data;

static void *thread_chunk(void *_d)
{
  chunk_data *d = _d;
  input *r = d->r;
  char *filename = d->filename;
  SNDFILE *in;
  SF_INFO in_info;
  chunk cur;
  float *buf;
  int i, j;

  in = open_audio_file(filename, &in_info);

  buf = malloc(sizeof(float) * 4096 * in_info.channels);
  if (buf == NULL) { fprintf(stderr, "out of memory\n"); return 0; }

  uint64_t nb_chunks = 0;
  while (in_info.frames) {
    sf_count_t count;
    sf_count_t to_read;
    to_read = in_info.frames > 4096 ? 4096 : in_info.frames;
    count = sf_read_float(in, buf, to_read * in_info.channels);
    if (count != to_read * in_info.channels) {
      fprintf(stderr, "error reading %s (%s)\n", filename, sf_strerror(in));
      return 0;
    }
    in_info.frames -= to_read;
    /* we suppose interleaved data */
    for (i = 0; i < in_info.channels; i++) {
      cur.min = buf[i];
      cur.max = buf[i];
      for (j = 1; j < to_read; j++) {
        float v = buf[i + j * in_info.channels];
        if (v < cur.min) cur.min = v;
        if (v > cur.max) cur.max = v;
      }
      r->chunks[nb_chunks] = cur;
      nb_chunks++;
    }
  }
printf("chunk nb_chunks %ld\n", nb_chunks);
  sf_close(in);
  free(buf);

  com_chunk_done();
  d->run = 0;
  return 0;
}

static void *thread_timer_chunk(void *_d)
{
  chunk_data *d = _d;
  while (d->run) {
    usleep(500*1000);
    if (!d->run) break;
    com_chunk_redraw();
  }
  free(_d);
  return 0;
}

void compute_chunks(input *r, char *filename)
{
  chunk_data *d;
  pthread_attr_t attr;
  pthread_t      tt;

  d = malloc(sizeof(chunk_data)); if (d == NULL) abort();
  d->r = r;
  d->filename = filename;
  d->run = 1;

  if (pthread_attr_init(&attr)) abort();
  if (pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED)) abort();
  if (pthread_create(&tt, &attr, thread_chunk, d)) abort();
  if (pthread_attr_destroy(&attr)) abort();

  if (pthread_attr_init(&attr)) abort();
  if (pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED)) abort();
  if (pthread_create(&tt, &attr, thread_timer_chunk, d)) abort();
  if (pthread_attr_destroy(&attr)) abort();
}

void load_chunk(input *r, char *audio_name, char *chunk_name)
{
  FILE *f;
  uint64_t chunk_count;

  if (chunk_name == NULL) { compute_chunks(r, audio_name); return; }

  chunk_count = (r->frames + 4095) / 4096 * r->channels;

  f = fopen(chunk_name, "r");
  if (f == NULL) {fprintf(stderr, "error opening %s\n", chunk_name); exit(1);}
  if (fread(r->chunks, sizeof(chunk), chunk_count, f) != chunk_count)
    { fprintf(stderr, "error reading %s\n", chunk_name); exit(1); }
  fclose(f);

  com_chunk_done();
}
