#include <X11/Xlib.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/select.h>
#include <unistd.h>
#include <time.h>
#include <string.h>
#include <math.h>
#include <pthread.h>
#include "audiotag.h"

#define BLACK 0
#define WHITE 1

typedef struct {
  Display *d;
  Window w;
  Pixmap dw;
  int width, height;
  GC colors[12];
  int must_paint;
  int must_compute;
  Font font;
} x_gui;

static void frame_to_time(char *s, input *r, int64_t frame)
{
  int64_t len_s = frame / r->samplerate;
  int64_t hour = len_s / 3600;
  int64_t minute = (len_s % 3600L) / 60;
  int64_t second = len_s % 60L;
  int64_t f = frame % r->samplerate;
  sprintf(s, "%2.2ld:%2.2ld:%2.2ld.%5.5ld", hour, minute, second, f);
}

static void put_display(x_gui *xg)
{
  XCopyArea(xg->d, xg->dw, xg->w, xg->colors[0],
            0, 0, xg->width, xg->height, 0, 0);
}

static void paint(x_gui *xg, gui *g, input *r, player *p)
{
  int i;
  int64_t start, inc, rinc;
  int chan;
  int x;
  char t[512];
  char t_start[64];
  char t_play[64];
  char t_head[64];
  char t_tail[64];
  int tlen;
  int dummy;
  int ascent;
  int descent;
  XCharStruct overall;
  int w_start, w_end;
  int s_start, s_end;
  int ls_start, ls_end;
  int64_t ns, ne;
  double xd;

  XFillRectangle(xg->d, xg->dw, xg->colors[BLACK],
                 0, 0, xg->width, xg->height);

  XDrawLine(xg->d, xg->dw, xg->colors[WHITE], 0, 100*r->channels+2,
            xg->width-1, 100*r->channels+2);

  XDrawLine(xg->d, xg->dw, xg->colors[WHITE], 0, 100*r->channels*2+6,
            xg->width-1, 100*r->channels*2+6);

  /* compute s_start/s_end the head/tail thing in high win */
  s_start = -1;
  s_end = -1;
  if (g->head != -1 && g->tail != -1 && g->head <= g->tail) {
    ns = g->head;
    ne = g->tail;
    s_start = ((double)ns / (double)r->frames) * (double)g->width;
    s_end = ((double)ne / (double)r->frames) * (double)g->width;
    if (s_end <= s_start) s_end = s_start + 1;
  }

  if (s_start != -1)
    XFillRectangle(xg->d, xg->dw, xg->colors[11],
                   s_start, 0, s_end-s_start, 100*r->channels+2);

  /* compute w_start/w_end the pixels in the high win where low win is */
  w_start = -1;
  w_end = -1;
  ns = g->start_sample;
  if (ns < 0) ns = 0;
  if (ns < r->frames) {
    ne = g->start_sample + g->samples_per_pixel * g->width;
    if (ne > 0 && ne > r->frames) ne = r->frames;
    if (ne > 0) {
      w_start = ((double)ns / (double)r->frames) * (double)g->width;
      w_end = ((double)ne / (double)r->frames) * (double)g->width;
      if (w_end <= w_start) w_end = w_start + 1;
    }
  }

  if (w_start != -1)
    XFillRectangle(xg->d, xg->dw, xg->colors[9],
                   w_start, 0, w_end-w_start, 100*r->channels+2);

  /* compute ls_start/ls_end the head/tail thing in low win */
  ls_start = -1;
  ls_end = -1;
  if (g->head != -1 && g->tail != -1 && g->head <= g->tail &&
      g->head < g->start_sample + g->samples_per_pixel * g->width &&
      g->tail >= g->start_sample) {
    ns = g->head;
    ne = g->tail;
    if (ns < g->start_sample) ns = g->start_sample;
    if (ne > g->start_sample + g->samples_per_pixel * g->width)
      ne = g->start_sample + g->samples_per_pixel * g->width;
    ls_start = (ns - g->start_sample) / g->samples_per_pixel;
    ls_end   = (ne - g->start_sample) / g->samples_per_pixel;
    if (ls_end <= ls_start) ls_end = ls_start + 1;
  }

  if (ls_start != -1)
    XFillRectangle(xg->d, xg->dw, xg->colors[11],
                   ls_start, 100*r->channels+3,
                   ls_end-ls_start, 100*r->channels+3);

  start = 0;
  inc = ceil((double)r->frames / xg->width);
  for (i = 0; i < xg->width; i++) {
    if (start < r->frames) {
      rinc = inc; if (start + rinc > r->frames) rinc = r->frames - start;
      for (chan = 0; chan < r->channels; chan++) {
        float min = -g->c[chan].all_min[i]*g->vzoom;
        float max = -g->c[chan].all_max[i]*g->vzoom;
        /* min/max reversed, but no care */
        if (min > 1) min = 1; if (min < -1) min = -1;
        if (max < -1) max = -1; if (max > 1) max = 1;
        XDrawLine(xg->d, xg->dw,
                  i >= w_start && i < w_end ? xg->colors[8] :
                  i >= s_start && i < s_end ? xg->colors[10] :
                    xg->colors[WHITE],
                  i, min*50+50+100*chan,
                  i, max*50+50+100*chan);
      }
    }
    start += inc;
  }

  start = g->start_sample;
  inc = g->samples_per_pixel;
  for (i = 0; i < xg->width; i++) {
    if (start >= 0 && start < r->frames) {
      rinc = inc; if (start + rinc > r->frames) rinc = r->frames - start;
      for (chan = 0; chan < r->channels; chan++) {
        float min = -g->c[chan].window_min[i]*g->vzoom;
        float max = -g->c[chan].window_max[i]*g->vzoom;
        /* min/max reversed, but no care */
        if (min > 1) min = 1; if (min < -1) min = -1;
        if (max < -1) max = -1; if (max > 1) max = 1;
        XDrawLine(xg->d, xg->dw,
                  i >= ls_start && i < ls_end ? xg->colors[10] :
                    xg->colors[WHITE],
                  i, min*50+54+100*chan+100*r->channels,
                  i, max*50+54+100*chan+100*r->channels);
      }
    }
    start += inc;
  }

  if (g->no_bars) goto no_bars;

  /* higher window start/play bars */
  x = R3(p->start, 0, r->frames, 0, g->width);
  XDrawLine(xg->d, xg->dw, xg->colors[2], x, 1, x, 100*r->channels);
  XDrawLine(xg->d, xg->dw, xg->colors[3], x-1, 1, x-1, 100*r->channels);
  XDrawLine(xg->d, xg->dw, xg->colors[3], x+1, 1, x+1, 100*r->channels);

  xd = R3(p->current, 0, r->frames, 0, g->width);
  if (xd >= 0 && xd < g->width) {
    x = xd;
    XDrawLine(xg->d, xg->dw, xg->colors[4], x, 1, x, 100*r->channels);
    XDrawLine(xg->d, xg->dw, xg->colors[5], x-1, 1, x-1, 100*r->channels);
    XDrawLine(xg->d, xg->dw, xg->colors[5], x+1, 1, x+1, 100*r->channels);
  }

  /* lower window start/play bars */
  xd = R3(p->start, g->start_sample, g->start_sample+g->samples_per_pixel,
                   0, 1);
  if (xd >= 0 && xd < g->width) {
    x = xd;
    XDrawLine(xg->d, xg->dw, xg->colors[2],
              x, 4+100*r->channels, x, 4+2*100*r->channels);
    XDrawLine(xg->d, xg->dw, xg->colors[3],
              x-1, 4+100*r->channels, x-1, 4+2*100*r->channels);
    XDrawLine(xg->d, xg->dw, xg->colors[3],
              x+1, 4+100*r->channels, x+1, 4+2*100*r->channels);
  }

  xd = R3(p->current, g->start_sample, g->start_sample+g->samples_per_pixel,
                   0, 1);
  if (xd >= 0 && xd < g->width) {
    x = xd;
    XDrawLine(xg->d, xg->dw, xg->colors[4],
              x, 4+100*r->channels, x, 4+2*100*r->channels);
    XDrawLine(xg->d, xg->dw, xg->colors[5],
              x-1, 4+100*r->channels, x-1, 4+2*100*r->channels);
    XDrawLine(xg->d, xg->dw, xg->colors[5],
              x+1, 4+100*r->channels, x+1, 4+2*100*r->channels);
  }

no_bars:
  /* text line */
  frame_to_time(t_start, r, p->start);
  frame_to_time(t_play, r, p->current);
  if (g->head != -1)
    frame_to_time(t_head, r, g->head);
  else
    sprintf(t_head, "None");
  if (g->tail != -1)
    frame_to_time(t_tail, r, g->tail);
  else
    sprintf(t_tail, "None");
  tlen = sprintf(t, "Start %s - Play %s - Head %s - Tail %s (%s) (vzoom %f)",
                 t_start, t_play, t_head, t_tail,
                 p->jack_error ? "JACK error" :
                   p->in_play ? "playing" : "not playing",
                 g->vzoom);
  XQueryTextExtents(xg->d, xg->font, t, tlen, &dummy, &ascent,
                    &descent, &overall);
  XDrawString(xg->d, xg->dw, xg->colors[WHITE], 2, r->channels*100*2+ascent+56,
              t, tlen);

  /* buttons */
  for (i = 0; i < g->bt_count; i++) {
    GC gc = g->bt[i].active ? xg->colors[7] : xg->colors[6];
    int bwidth = g->bt[i].width;
    int bheight = g->bt[i].height;
    int bx = g->bt[i].x;
    int by = g->bt[i].y;
    XFillRectangle(xg->d, xg->dw, gc, bx, by, bwidth, bheight);
    XQueryTextExtents(xg->d, xg->font, g->bt[i].t, g->bt[i].tlen,
                      &dummy, &ascent, &descent, &overall);
    XDrawString(xg->d, xg->dw, xg->colors[BLACK], bx+(bwidth-overall.width)/2,
                by+(bheight+ascent+descent)/2-descent,
                g->bt[i].t, g->bt[i].tlen);
  }
}

void run_gui(gui *g, input *r, player *p)
{
  int i;
  x_gui xg;
  XGCValues gcv;
  XWindowAttributes att;
  XColor col[12];
  int xfd;
  struct timespec last_paint_time;
  struct timespec cur_time;
  struct timespec diff;
  uint64_t refresh_interval;
  FILE *region;

  region = NULL; /* shut up gcc */

  if (g->region_filename != NULL) {
    region = fopen(g->region_filename, "a");
    if (region == NULL) { perror(g->region_filename); exit(1); }
  }

  xg.width = g->width;
  xg.height = g->height;
  refresh_interval = 1000000000 / g->refresh_rate;  /* nanosecond */

  g->samples_per_pixel = 4096;
  g->start_sample = -100*4096;

  g->head = -1;
  g->tail = -1;

  g->bt[0].t = "head";
  g->bt[0].tlen = 4;
  g->bt[0].x = 4;
  g->bt[0].y = r->channels*100*2+12;
  g->bt[0].width = 100;
  g->bt[0].height = 30;
  g->bt[0].active = 0;

  g->bt[1].t = "tail";
  g->bt[1].tlen = 4;
  g->bt[1].x = 4 + 110;
  g->bt[1].y = r->channels*100*2+12;
  g->bt[1].width = 100;
  g->bt[1].height = 30;
  g->bt[1].active = 0;

  g->bt[2].t = "add";
  g->bt[2].tlen = 3;
  g->bt[2].x = 4 + 110*2;
  g->bt[2].y = r->channels*100*2+12;
  g->bt[2].width = 100;
  g->bt[2].height = 30;
  g->bt[2].active = 0;

  g->bt[3].t = "zoom reset";
  g->bt[3].tlen = 10;
  g->bt[3].x = 4 + 110*3;
  g->bt[3].y = r->channels*100*2+12;
  g->bt[3].width = 100;
  g->bt[3].height = 30;
  g->bt[3].active = 0;

  g->bt[4].t = "quit";
  g->bt[4].tlen = 4;
  g->bt[4].x = g->width - 104;
  g->bt[4].y = r->channels*100*2+12;
  g->bt[4].width = 100;
  g->bt[4].height = 30;
  g->bt[4].active = 0;

  g->bt_count = 5;

  g->vzoom = 1;
  /* g->no_bars set in main */

  xg.d = XOpenDisplay(0);
  if (xg.d == NULL) { fprintf(stderr, "X error\n"); exit(1); }

  for (i = 0; i < 12; i++) {
    XColor rcol;
    if (!XAllocNamedColor(xg.d, DefaultColormap(xg.d, DefaultScreen(xg.d)),
                          g->colors[i], &col[i], &rcol))
      { fprintf(stderr, "error alloc color '%s'\n", g->colors[i]); exit(1); }
  }

  xg.w = XCreateSimpleWindow(xg.d, DefaultRootWindow(xg.d), 0, 0,
           xg.width, xg.height, 0, 0, col[BLACK].pixel);
  XSelectInput(xg.d, xg.w, ButtonPressMask | ButtonReleaseMask | ExposureMask);
  XMapWindow(xg.d, xg.w);

  XGetWindowAttributes(xg.d, xg.w, &att);
  xg.dw = XCreatePixmap(xg.d, xg.w, xg.width, xg.height, att.depth);

  xg.font = XLoadFont(xg.d, g->font);

  for (i = 0; i < 12; i++) {
    xg.colors[i] = XCreateGC(xg.d, xg.w, 0, NULL);
    XCopyGC(xg.d, DefaultGC(xg.d, DefaultScreen(xg.d)), -1L, xg.colors[i]);
    gcv.foreground = col[i].pixel;
    gcv.font = xg.font;
    XChangeGC(xg.d, xg.colors[i], GCForeground | GCFont, &gcv);
  }

  xfd = ConnectionNumber(xg.d);

  xg.must_paint = 1;

  com_compute_lo_responsive();

  xg.must_compute = 0;

  while (1) {
    fd_set rset;
    XEvent ev;
    int must_draw = 0;

    while (XCheckMaskEvent(xg.d, -1L, &ev)) {
      switch(ev.type) {
      case Expose: must_draw = 1; break;
      case ButtonPress: {
        /* shift+button 4 = vzoom inc */
        if (ev.xbutton.button == 4 && (ev.xbutton.state & ShiftMask)) {
          g->vzoom *= 1.2;
          xg.must_paint = 1;
          break;
        } else
        /* shift+button 5 = vzoom dec */
        if (ev.xbutton.button == 5 && (ev.xbutton.state & ShiftMask)) {
          g->vzoom /= 1.2;
          xg.must_paint = 1;
          break;
        } else
        if (ev.xbutton.button == 3) {
          int ip = GET(p->in_play);
          if (ip) stop_play(p); else start_play(p);
          xg.must_paint = 1;
        } else
        if (ev.xbutton.button == 1) {
          /* higher sample zone */
          if (ev.xbutton.y < 100*r->channels &&
              !p->in_play) {
            double start = R3(ev.xbutton.x, 0, g->width, 0, r->frames);
            if (start < 0) start = 0;
            if (start > r->frames-1) start = r->frames-1;
            p->start = start;
            xg.must_paint = 1;
          }
          /* lower sample zone */
          if (ev.xbutton.y >= 4+100*r->channels &&
              ev.xbutton.y <= 4+2*100*r->channels &&
              !p->in_play) {
            int64_t start = R3(ev.xbutton.x, 0, 1, g->start_sample,
                            g->start_sample+g->samples_per_pixel);
            if (start < 0) start = 0;
            if (start > r->frames-1) start = r->frames-1;
            p->start = start;
            xg.must_paint = 1;
          }
          /* one of the buttons */
          for (i = 0; i < g->bt_count; i++)
            if (ev.xbutton.x >= g->bt[i].x &&
                ev.xbutton.x <= g->bt[i].x+g->bt[i].width &&
                ev.xbutton.y >= g->bt[i].y &&
                ev.xbutton.y <= g->bt[i].y+g->bt[i].height) {
              g->bt[i].active = 1;
              xg.must_paint = 1;
              break;
            }
        } else
        /* button 4: scroll up (zoom in in low window, scroll in high) */
        if (ev.xbutton.button == 4) {
          if (ev.xbutton.y <= 100*r->channels) {
            g->start_sample += g->samples_per_pixel * g->width/10;
          } else {
            /* zoom in (less samples per pixel) */
            int64_t spp = g->samples_per_pixel;
            int64_t nspp = (float)spp / 1.2;
            if (nspp == g->samples_per_pixel) nspp = spp - 1;
            if (nspp == 0) nspp = 1;
            g->samples_per_pixel = nspp;
            g->start_sample += ev.xbutton.x * (spp - nspp);
          }
          xg.must_compute = 1;
          //xg.must_paint = 1;
        } else
        /* button 5: scroll down (zoom out in low win, scroll in high) */
        if (ev.xbutton.button == 5) {
          if (ev.xbutton.y <= 100*r->channels) {
            g->start_sample -= g->samples_per_pixel * g->width/10;
          } else {
            /* zoom out (more samples per pixel) */
            int64_t spp = g->samples_per_pixel;
            int64_t nspp = (float)spp * 1.2;
            if (nspp == g->samples_per_pixel) nspp = spp + 1;
            if (nspp > r->frames) nspp = r->frames;
            g->samples_per_pixel = nspp;
            g->start_sample += ev.xbutton.x * (spp - nspp);
          }
          xg.must_compute = 1;
          //xg.must_paint = 1;
        }
        break;
      }
      case ButtonRelease: {
        if (ev.xbutton.button == 1) {
          for (i = 0; i < g->bt_count; i++)
            if (ev.xbutton.x >= g->bt[i].x &&
                ev.xbutton.x <= g->bt[i].x+g->bt[i].width &&
                ev.xbutton.y >= g->bt[i].y &&
                ev.xbutton.y <= g->bt[i].y+g->bt[i].height) {
              if (!g->bt[i].active) break;
              switch (i) {
              case 0: g->head = p->start; break;
              case 1: g->tail = p->start; break;
              case 2: {
                char h[64];
                char t[64];
                if (g->head != -1 && g->tail != -1) {
                  struct timespec curt;
                  struct tm *tl;
                  clock_gettime(CLOCK_REALTIME, &curt);
                  tl = localtime(&curt.tv_sec);
                  frame_to_time(h, r, g->head);
                  frame_to_time(t, r, g->tail);
                  if (g->region_filename != NULL)
                    if (fprintf(region,
                "REGION %s %s (saved %d %2.2d %2.2d T %2.2d %2.2d %2.2d)\n",
                           h, t, tl->tm_year+1900, tl->tm_mon+1, tl->tm_mday,
                           tl->tm_hour, tl->tm_min, tl->tm_sec) < 0) {
                      perror(g->region_filename);
                      exit(1);
                    }
                  g->head = g->tail = -1;
                }
                break;
              }
              case 3:
                g->samples_per_pixel = 4096;
                g->start_sample = -100*4096;
                xg.must_compute = 1;
                g->vzoom = 1;
                break;
              case 4:
                if (g->region_filename != NULL)
                  if (fclose(region)) perror(g->region_filename); exit(0);
              }
              break;
            }
          for (i = 0; i < g->bt_count; i++) g->bt[i].active = 0;
          xg.must_paint = 1;
        }
      }
      default: break;
      }
    }
    if (xg.must_compute) com_compute_lo_responsive();
    if (xg.must_paint) {
      paint(&xg, g, r, p);
      clock_gettime(CLOCK_REALTIME, &last_paint_time);
    }
    if (xg.must_paint || must_draw) put_display(&xg);
    xg.must_paint = 0;
    xg.must_compute = 0;

    XFlush(xg.d);

    FD_ZERO(&rset);
    FD_SET(xfd, &rset);
    FD_SET(gui_read, &rset);
    if (select(xfd > gui_read ? xfd+1 : gui_read+1, &rset, NULL, NULL,
               NULL) == -1) abort();

    if (FD_ISSET(gui_read, &rset)) {
      char c[4096];
      int len;
      len = read(gui_read, c, 4096);
      for (i = 0; i < len; i++) {
        if (c[i] == 1) { xg.must_paint = 1; }
      }
      clock_gettime(CLOCK_REALTIME, &cur_time);
      if (cur_time.tv_nsec < last_paint_time.tv_nsec) {
        diff.tv_sec = cur_time.tv_sec - last_paint_time.tv_sec - 1;
        diff.tv_nsec = 1000000000 + cur_time.tv_nsec - last_paint_time.tv_nsec;
      } else {
        diff.tv_sec = cur_time.tv_sec - last_paint_time.tv_sec;
        diff.tv_nsec = cur_time.tv_nsec - last_paint_time.tv_nsec;
      }
      if (diff.tv_sec || diff.tv_nsec >= refresh_interval) {
        /* force p->current in view if in_play is on and !no_follow */
        if (!g->no_follow && p->in_play) {
          if (p->current < g->start_sample ||
              p->current >= g->start_sample +
                  (g->width*9/10) * g->samples_per_pixel) {
            g->start_sample = p->current - (g->width/10) *
                g->samples_per_pixel;
            xg.must_compute = 1;
          }
        }
        xg.must_paint = 1;
      }
    }
  }
}
