#include <stdio.h>
#include <sndfile.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sndfile.h>
#include "audiotag.h"

void alloc_mem(gui *g, input *r, player *p)
{
  int i;
  uint64_t chunk_count;
  SF_INFO inf;
  SNDFILE *in = open_audio_file(p->audio_file, &inf);
  sf_close(in);

  r->frames = inf.frames;
  r->channels = inf.channels;
  r->samplerate = inf.samplerate;

  chunk_count = (r->frames + 4095) / 4096 * r->channels;
  r->chunks = malloc(chunk_count * sizeof(chunk));
  if (r->chunks == NULL) { perror("malloc"); exit(1); }
  for (i = 0; i < chunk_count; i++) {
    r->chunks[i].min = -0.5;
    r->chunks[i].max = 0.5;
  }

  g->c = malloc(sizeof(channel) * r->channels);
  if (g->c == NULL) { perror("malloc"); exit(1); }

  for (i = 0; i < r->channels; i++) {
    int j;
    g->c[i].window_min = malloc(sizeof(float) * g->width);
    if (g->c[i].window_min == NULL) { perror("malloc"); exit(1); }
    for (j = 0; j < g->width; j++) g->c[i].window_min[j] = -0.5;
    g->c[i].window_max = malloc(sizeof(float) * g->width);
    if (g->c[i].window_max == NULL) { perror("malloc"); exit(1); }
    for (j = 0; j < g->width; j++) g->c[i].window_max[j] = 0.5;
    g->c[i].all_min = malloc(sizeof(float) * g->width);
    if (g->c[i].all_min == NULL) { perror("malloc"); exit(1); }
    for (j = 0; j < g->width; j++) g->c[i].all_min[j] = -0.5;
    g->c[i].all_max = malloc(sizeof(float) * g->width);
    if (g->c[i].all_max == NULL) { perror("malloc"); exit(1); }
    for (j = 0; j < g->width; j++) g->c[i].all_max[j] = 0.5;
  }
}

int force_format = 0;
int force_channels = 1;
int force_samplerate = 48000;
float timer_length = 50;

void *open_audio_file(char *f, void *inf)
{
  SNDFILE *in;
  SF_INFO *in_info = inf;
  SF_INFO _inf;
  if (inf == NULL) in_info = &_inf;
  in_info->samplerate = force_samplerate;
  in_info->channels = force_channels;
  in_info->format = force_format;
  in_info->sections = 1;
  in_info->seekable = 1;
  in = sf_open(f, SFM_READ, in_info);
  if (in == NULL) { fprintf(stderr, "error opening %s: %s\n", f, sf_strerror(NULL)); exit(1); }
  return in;
}

void usage(void)
{
  printf("Usage: plot [options] <audio file>\n");
  printf("Options:\n");
  printf("    -width <width>\n");
  printf("        set width of window\n");
  printf("    -height <height>\n");
  printf("        set height of window\n");
  printf("    -fps <frames per second>\n");
  printf("        set frame rate when playing audio (1..100; default 10)\n");
  printf("    -font <font>\n");
  printf("        set font (default \"-*-fixed-medium-r-*-*-12-*\")\n");
  printf("    -color <color number> <color name (eg. \"#fff\")>\n");
  printf("        set color (color number in [0..11])\n");
  printf("    -no-follow\n");
  printf("        don't follow the playing position if it goes out of view\n");
  printf("    -cf <chunk file>\n");
  printf("        use given file as precomputed chunk file\n");
  printf("        if no chunk file is used, computation done at runtime\n");
  printf("    -rf <region file>\n");
  printf("        use given file for region file\n");
  printf("        if no regin file given, button 'add' does nothing\n");
  printf("    -no-jack\n");
  printf("        don't use the jack stuff\n");
  printf("    -jack-connect <to port (eg. system:playback_1)>\n");
  printf("        connect jack out port (may appear more than once)\n");
  printf("    -no-bars\n");
  printf("        don't display vertical bars (head, tail, pos, ...)\n");
  printf("    -raw <raw format>\n");
  printf("        open as raw file (samplerate/channels to be set too)\n");
  printf("        format one of: s8, u8, s16le, floatle\n");
  printf("    -c <channel count (default 1)>\n");
  printf("        used only if -f is given too\n");
  printf("    -r <samplerate (default 48000)>\n");
  printf("        used only if -f is given too\n");
  printf("    -tl <timer value (default 50, unit ms)>\n");
  printf("        the lower the more responsive, but the more flickering\n");
  exit(1);
}

int main(int n, char **v)
{
  gui g;
  input r;
  player p;
  int i;
  char *audio_file = NULL;
  char *chunk_file = NULL;
  char *region_file = NULL;
  int no_jack = 0;

  g.colors[0] = "black";
  g.colors[1] = "#aaa";
  g.colors[2] = "#f44";
  g.colors[3] = "#944";
  g.colors[4] = "#2f2";
  g.colors[5] = "#494";
  g.colors[6] = "#ccc";
  g.colors[7] = "#66a";
  g.colors[8] = "#ccf"; /* low window position in high window foreground */
  g.colors[9] = "#99c"; /* low window position in high window background */
  g.colors[10] = "#dda";
  g.colors[11] = "#aa7";
  g.font = "-*-fixed-medium-r-*-*-12-*";
  //g.font = "-*-fixed-medium-r-*-*-20-*";
  g.refresh_rate = 10;

  g.width = 1204;
  g.height = 480;
  g.no_follow = 0;
  g.no_bars = 0;

  for (i = 1; i < n; i++) {
    if (!strcmp(v[i], "-color")) {
      int col;
      if (i >= n-2) usage();
      col = atoi(v[i+1]); if (col < 0 || col > 11) usage();
      g.colors[col] = v[i+2];
      i += 2;
      continue;
    }
    if (!strcmp(v[i], "-font")) {
      if (i >= n-1) usage();
      g.font = v[i+1];
      i++;
      continue;
    }
    if (!strcmp(v[i], "-fps")) {
      int r;
      if (i >= n-1) usage();
      g.refresh_rate = r = atoi(v[i+1]);
      if (g.refresh_rate < 1) g.refresh_rate = 1;
      if (g.refresh_rate > 100) g.refresh_rate = 100;
      if (r != g.refresh_rate)
        fprintf(stderr, "WARNING: setting fps to %d but you asked %s\n",
                g.refresh_rate, v[i+1]);
      i++;
      continue;
    }
    if (!strcmp(v[i], "-width")) {
      if (i >= n-1) usage();
      g.width = atoi(v[i+1]);
      if (g.width < 1) g.width = 1; if (g.width > 32767) g.width = 32767;
      i++;
      continue;
    }
    if (!strcmp(v[i], "-height")) {
      if (i >= n-1) usage();
      g.height = atoi(v[i+1]);
      if (g.height < 1) g.height = 1; if (g.height > 32767) g.height = 32767;
      i++;
      continue;
    }
    if (!strcmp(v[i], "-no-follow")) {
      g.no_follow = 1;
      continue;
    }
    if (!strcmp(v[i], "-cf")) {
      if (i >= n-1) usage();
      if (chunk_file != NULL) usage();
      chunk_file = v[i+1];
      i++;
      continue;
    }
    if (!strcmp(v[i], "-rf")) {
      if (i >= n-1) usage();
      if (region_file != NULL) usage();
      region_file = v[i+1];
      i++;
      continue;
    }
    if (!strcmp(v[i], "-no-jack")) {
      no_jack = 1;
      continue;
    }
    if (!strcmp(v[i], "-jack-connect")) {
      extern const char *jack_connection[128];
      extern int jack_connect_count;
      if (i >= n-1) usage();
      if (jack_connect_count == 128) usage();
      jack_connection[jack_connect_count] = v[i+1];
      jack_connect_count++;
      i++;
      continue;
    }
    if (!strcmp(v[i], "-no-bars")) {
      g.no_bars = 1;
      continue;
    }
    if (!strcmp(v[i], "-raw")) {
      if (i >= n-1) usage();
      force_format = SF_FORMAT_RAW;
      if      (!strcmp(v[i+1], "s8"))
        force_format |= SF_FORMAT_PCM_S8;
      else if (!strcmp(v[i+1], "u8"))
        force_format |= SF_FORMAT_PCM_U8;
      else if (!strcmp(v[i+1], "s16le"))
        force_format |= SF_FORMAT_PCM_16 | SF_ENDIAN_LITTLE;
      else if (!strcmp(v[i+1], "floatle"))
        force_format |= SF_FORMAT_FLOAT | SF_ENDIAN_LITTLE;
      else usage();
      i++;
      continue;
    }
    if (!strcmp(v[i], "-c")) {
      if (i >= n-1) usage();
      force_channels = atoi(v[i+1]);
      i++;
      continue;
    }
    if (!strcmp(v[i], "-r")) {
      if (i >= n-1) usage();
      force_samplerate = atoi(v[i+1]);
      i++;
      continue;
    }
    if (!strcmp(v[i], "-tl")) {
      if (i >= n-1) usage();
      timer_length = atof(v[i+1]);
      i++;
      continue;
    }
    if (audio_file == NULL) { audio_file = v[i]; continue; }
    usage();
  }

  if (audio_file == NULL) usage();

  g.region_filename = region_file;
  p.audio_file = audio_file;
  p.start = 0;
  p.no_jack = no_jack;

  init_com();
  alloc_mem(&g, &r, &p);

  run_compute(&g, &r, &p);
  init_jack(&p);
  load_chunk(&r, audio_file, chunk_file);
  run_gui(&g, &r, &p);

  return 0;
}
